//------Standard C/C++ libraries
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stream.h>

//------Performer stuff
#include <Performer/pf/pfChannel.h>
#include <Performer/pf/pfEarthSky.h>
#include <Performer/pf/pfLightSource.h>
#include <Performer/pf/pfDCS.h>
#include <Performer/pf/pfText.h>
#include <Performer/pfdu.h>
#include <Performer/pf/pfGeode.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pr/pfMaterial.h>
//#include <Performer/pr.h>
#include <Performer/pr/pfString.h>
#include <Performer/pr/pfFont.h>
//------Cave stuff
#include <pfcave.h>

//------Local Includes
#include "FlexArrayTmpl.H"
#include "pfFlexArrayTmpl.H"
#include "ReadAtomFiles.hh" // to read ascii datafiles from sim
#include "CavePointer.hh"
#include "NavBase.hh"
#include "Label3D.hh"

//--------Local defines
#ifndef THRESH
// Threshold is used by the joystick to prevent drifting
// (the joystick magnitude is never actually 0).
#define THRESH(val,thresh) (val>thresh || val<-thresh)
#endif
// Typedefs
typedef pfDCS *pfDCSPtr;
//------Global decls
static pfGeoState *NoLightGState;
//------Utility subroutines for creating performer scene graph
pfNode *NewScene(void); // just creates a cone marking x=0,y=0,z=0 location
// Creates a new default atom (if a custom atom file doesn't already exist).
pfDCS *NewAtom(float red, float green, float blue, float transparency,int refinement=32);
inline pfDCS *NewAtom(float rgba[4],int refinement=32){
  return NewAtom(rgba[0],rgba[1],rgba[2],rgba[3],refinement);
}
// Initialize the Performer/Cave environment.  Called once at startup
// before any performer data structures are allocated.
void StartCave(int *argc,char *argv[],pfDCSPtr &movingdcs,pfDCSPtr &stationarydcs);
inline void UpdateCave(void); // called in main event loop
void StopCave(void);
//---Default Atom Colors----{r,  g, b, a}-----------------------
float DefaultAtomColor[6][4] = {{0.8,0.0,0.2,1.0},  // reddish atom
				{0.2,0.0,0.8,1.0},  // bluish atom
				{0.6,0.5,0.1,1.0},  // yellowish atom
				{1.0,0.0,1.0,1.0},  // bright purple atom
				{1.0,1.0,0.0,1.0},  // bright orange atom
				{0.0,1.0,0.0,1.0}}; // bright green atom

// This entire class handle the reading and the storage of
// commandline arguments for use within the program.
struct CommandLine {
  char *programname;
  char *basefilename;
  int minindex,maxindex;
  enum CaveType {Desk,Cave,Wall,Simulator};
  CaveType cave_type;  // type of cave you are displaying in.
  int use_default_atoms; // force application to use internally generated atom geometries
  int refinement; // refinement for default atom geometries

  char filenamebuffer[64];
  
  void usage(){
    printf("Usage: %s -file <basefilename> -range <minfilenumber>:<maxfilenumber>\n",programname);
    puts("where <basefilename> is the root name of a file for a sequence of files of");
    puts("the form <basefilename>.<filenumber>.  The range defines the minimum filenumber");
    puts("and the maximum for <filenumber>.  The program will then read in all files");
    puts("(from min to max filenumber) and interpolate linearly between the datasets");
    puts("for time-animations of the data.");
    puts("Other Optional Flags:");
    puts("-cave, -desk, -wall, -simulator:  Each of these flags defines a different");
    puts("\tdisplay device that the program is running on (since it cant automatically");
    puts("\tsense the the differences between and Idesk and an Cave.  So the -cave flag");
    puts("\ttells it to adjust its user-interface geometry for display inside of a cave");
    puts("\tand the -desk flag adjusts the user-interface geometry for the immersadesk.");
    puts("\tThe default is -cave.");
    puts("\n-default: Forces the program to use internally created geometries to ");
    puts("\trepresent the atoms.  The internal geometry is a sphere, but you can");
    puts("\tselect the resolution of the sphere (number of facets) using the");
    puts("\t-refinement option.  Using fewer facets created more jagged spheres, but");
    puts("\tless complicated scenes can be drawn faster in 3D.");
    puts("\n\tIf the default (internal) geometries are not used, then the program");
    puts("\twill search for custom geometries in the current working directory.");
    puts("\tThe custom geometries have the same name as the atom they represent (as");
    puts("\tprinted in the xyz files) and have the extension .vrml, .wrl, or .iv");
    puts("\tdepending on the type of geometry they contain.  If a custom geometry");
    puts("\tcannot be found, then it will fall back on the default (internal)");
    puts("\tgeometries.");
    puts("\n-refinement: This sets the number of faces used to create a sphere");
    puts("\twhen the internal (default) geometries are used.  Fewer faces means");
    puts("\tfaster drawing since the objects are less complicated.  However, it");
    puts("\talso means the spheres are less smooth looking.  The default is 32.");
  }
  void setDefaults(){
    programname="";
    basefilename="fort";
    minindex=maxindex=-1;
    cave_type=Cave;
    use_default_atoms=0;
    refinement = 24;
  }

  CommandLine(){
    setDefaults();
  }

  CommandLine(int argc,char *argv[]){
    setDefaults();
    parse(argc,argv);
  }

  // Read the commandline arguments.
  void parse(int argc,char *argv[]){
    programname=argv[0];
    for(int i=1;i<argc;i++){
      char *str=argv[i];
      if(*str=='-'){
	char *val=argv[i+1];
	str++;
	if(!strcmp(str,"desk"))          {cave_type=Desk;     }
	else if(!strcmp(str,"cave"))     {cave_type=Cave;     }
	else if(!strcmp(str,"wall"))     {cave_type=Wall;     }
	else if(!strcmp(str,"simulator")){cave_type=Simulator;}
	else if(!strcmp(str,"file"))     {basefilename=val;   i++;}
	else if(!strcmp(str,"range"))    {sscanf(val,"%u:%u",&minindex,&maxindex);}
	else if(!strcmp(str,"defgeom"))  {use_default_atoms=1; }
	else if(!strcmp(str,"default"))  {use_default_atoms=1; }
	else if(!strcmp(str,"refine") || !strcmp(str,"refinement"))
	                                 { refinement=atoi(val); }
      }
    }
    if(minindex==-1 || maxindex==-1) // if not properly initialized
      prompt(); // get it from the commandline
  }

  void prompt(){
    char buffer[256];
    usage();
    puts("Since you didn't use the commandline arguments to set these parameters,");
    puts("I am going to prompt you for them.\n\n");
    puts("The filenames are of the format <basename>.<number> where the basename");
    puts("is something like \"fort\" and the <number> ranges from 1 to 100 for");
    puts("a sequence of 100 files named \"fort.1\" to \"fort.100\"");
    puts("Press return to select the default.");
    printf("What is the basename for the datafile (default is [%s])\n",basefilename);
    gets(filenamebuffer); // read a line
    if(*filenamebuffer!='\0') // if it is nonempty
      basefilename=filenamebuffer; // then it is the new name for the basefilename
    printf("Base filename is [%s]\n",basefilename); // inform the user of the selected name
    printf("\nWhat is the minimum and the maximum filenumber for these datafiles (no default)\n");
    printf("(use the form minimum:maximum)\n");
    gets(buffer);
    sscanf(buffer,"%u:%u",&minindex,&maxindex);
    printf("Range for file numbers is %u:%u\n",minindex,maxindex);
    printf("\nIs this running on a Cave, Immersadesk,Iwall, or Cave Simulator\n");
    printf("(Choices are {cave,desk,wall,sim}.... The default is [cave])");
    gets(buffer);
    switch(*buffer){
    case 'd':
      puts("Selected [desk]");
      cave_type = Desk;
      break;
    case 'w':
      puts("Selected [wall]");
      cave_type = Wall;
      break;
    case 's':
      puts("Selected [sim]");
      cave_type=Simulator;
      break;
    case '/0':
    case 'c':
      puts("Selected [cave]");
      cave_type=Cave;
      break;
    default:
      printf("Didn't understand your choice [%s].  So I'm selecting [cave].\n",buffer);
      break;
    }
    puts("Do you want to use custom geometries (internally generated spheres).");
    puts("or do you have some custom geometries of the form <atomname>.wrl or");
    puts(".iv or .vrml that you would like to use? (yes or no.  Default is [yes])");
    gets(buffer);
    if(*buffer=='n') {
      use_default_atoms=1;
      puts("Using default (internally generated) atoms.");
    }
    else {
      puts("Using custom atoms and falling back to default (internally generated) atoms if");
      puts("a custom one cannot be found.");
    }
    puts("For the default (internally generated) geometries, what level of refinement");
    puts("do you want to use.  (Default is [32]).");
    gets(buffer);
    if(*buffer!='\0')
      refinement=atoi(buffer);
    printf("Selected a refinement of %u\n",refinement);
    puts("\nThanks for your input... now lets start the program\n\n");
  }
};

int main(int argc,char *argv[]){ 
  //**************Local Vars
  int i;
  pfDCS *movingDCS;
  pfDCS *stationaryDCS;
  FlexArray<AtomDef> atomlist;
  CommandLine cmdln(argc,argv); // read and process the commandline parameters

  // ***********Start the Cave Processes********************
  StartCave(&argc,argv,movingDCS,stationaryDCS);

  // ***********Setup Root-Level DCS's***********************
  pfDCS *modelDCS;
  FlexArray<pfDCS*> dcslist;
  
  NavFly *navigation= new NavFly(movingDCS); // use moving DCS for navigation (flying around)
  modelDCS = new pfDCS; // DCS that contains all of the atoms
  movingDCS->addChild(modelDCS); // model dcs is a child of the moving DCS.
  modelDCS->setScale(1.0); // Set scale initially to 1.0
  
  // ***********Load up the xyz datafiles***********************
  TimeAtomList timeAtomList(cmdln.basefilename,cmdln.minindex,cmdln.maxindex); 
  timeAtomList.getAtoms(0.0,atomlist); // select the first dataset initially
  // Now lets create one geometric representation per atom type
  dcslist.setSize(atomlist.getSize());
  // Now we need to get a list of atom DCS's
  // For each type of atom, we either load in a custom atom geometry
  // from disk (if it can be found and the -default flag was *not* used).
  // Or we use NewAtom() to generate a custom geometry.
  //
  // In the future, you might want to have a separate geometry generated
  // for each atom.  currently a given geometry is shared for all atoms
  // that are of the same type.
  FlexArray<pfDCS *> atomGeoms;
  printf("Number of atom types = %u\n",timeAtomList.atom_types.nTypes());
  for(i=0;i<timeAtomList.atom_types.nTypes();i++){
    char buffer[256];// filename 
    typedef char *str;
    str extensions[3]={"iv","wrl","vrml"};
    pfDCS *atom=0;
    // if a .wrl or .iv file exists for this atom type, use it!
    pfNode *geom=0;
    // try to load custom atoms unless explictly told otherwise
    if(!cmdln.use_default_atoms){ 
      for(int j=0;j<3 && !atom;j++){ // for each extension type
	sprintf(buffer,"%s.%s",timeAtomList.atom_types.getTypeName(i),extensions[j]);
	if((geom=pfdLoadFile(buffer))){
	  // only do this if geometry was successfully loaded from file
	  atom = new pfDCS;
	  atom->addChild(geom);
	  break; // successful... exit loop
	}
      }
    }
    // if all attempts to load custom atom geometries fails...
    if(!atom) // (no atom geometry files) create a default atom geoms
      atom=NewAtom(DefaultAtomColor[i],cmdln.refinement); // create a new atom geometry
    atomGeoms.append(atom); // append the new atom geometry to the list
  }

  // For each atom, assign the proper atom geometry and location based
  // on the atoms type and position in the lattice.
  for(i=0;i<atomlist.getSize();i++){
    pfDCS *tmpdcs;
    int atomtype;
    // create a DCS for each atom so they can be moved independently
    dcslist[i] = tmpdcs = new pfDCS;
    // add that DCS to the main model DCS
    modelDCS->addChild(tmpdcs); 
    atomtype = atomlist[i].type;
    tmpdcs->addChild(atomGeoms[atomtype]); // assign geometry based on atom type
    // assign the location of the geometry based on the atom's position
    tmpdcs->setTrans(atomlist[i].p.cart.x,atomlist[i].p.cart.y,atomlist[i].p.cart.z);
  }

  // *********Now we create labels to identify the mode for the user interface ****
  // set position for the label in the upper lefthand corner of the display.
  // This is different depending on whether you are using a Cave, Wall, or Idesk.
  pfMatrix mat;
  switch(cmdln.cave_type){
  case CommandLine::Wall:
    puts("Configure for the Wall");
    mat.makeIdent();
    mat.preTrans(-4.6,0.0,8.3,mat); // Move to upper lefthand corner of the Wall
    break;
  case CommandLine::Desk:
    puts("Configure for the Idesk");
    mat.makeRot(-40.19,1.0,0.0,0.0);
    mat.preTrans(-3.79,0.0,8.0,mat); // Move to upper lefthand corner of the Idesk
    break;
  case CommandLine::Simulator:
  case CommandLine::Cave:
  default:
    puts("Configure for the cave");
    mat.makeIdent();
    mat.preTrans(-4.0,5.0,8.0,mat); // Move to upper lefthand corner of the cave/simulator
    break;
  }
  pfSCS *labelscs = new pfSCS(mat);
  mat.preTrans(0.0,0.0,-0.3,mat); // Now move just below the Mode label
  pfSCS *valscs = new pfSCS(mat);
  stationaryDCS->addChild(labelscs); // this is put in the non-moving DCS (non-flying)
  stationaryDCS->addChild(valscs);
  // Label3D objects manage the display of text inside of the Cave using Performer
  Label3D *mode_label = new Label3D(labelscs);
  Label3D *val_label = new Label3D(valscs);
  mode_label->setFontSize(0.3); // Font Height is .3 feet
  val_label->setFontSize(0.3);
  mode_label->setLabel("Mode: (press any wand button to change modes)"); // Initial mode label
  val_label->setLabel("<>"); // Initial value label
  // Speed controls how responsive the joystick is.
#define SPEED 0.1L
  // These defines identify the different modes of operation for the user interface
#define MODE_INITIALIZE -1
#define MODE_SCALE_MODEL 0
#define MODE_SCALE_ATOMS 1
#define MODE_NAVIGATE 2
#define MODE_CHANGE_TIME 3
  while (!CAVEgetbutton(CAVE_ESCKEY)) {
    //--State Variables
    static double s=1.0,sa=1.0;
    static float mintime=-1.0,maxtime=-1.0,t=0.0;
    static int mode=MODE_INITIALIZE,button_was_down=0;
    //--Temporary variables
    char buffer[32];
    double jx=CAVE_JOYSTICK_X; // joystick left-to-right movement (-1 to 1)
    double jy=CAVE_JOYSTICK_Y; // joystick forward-to-backward movement (-1 to 1)
    
    UpdateCave(); // Update the performer/cave display processes

    //---Change mode of operation if any button was pressed
    if(CAVEBUTTON1||CAVEBUTTON2||CAVEBUTTON3){
      if(!button_was_down){ // and the button was *up* on previous cycle.
	button_was_down=1;
	if(CAVEBUTTON1) {
	  mode--; // if leftbutton pressed... decrement modenum
	  if(mode<0) mode=3;
	}
	else mode++; //if middle or right button is pressed... increment mode
	mode %= 4; // increment mode and then modulo # of modes
	switch(mode){// handle new mode
	case MODE_SCALE_MODEL:
	  puts("Set Mode: Scale Model");
	  mode_label->setLabel("Mode: Scale Model");
	  sprintf(buffer,"scale=%5.2lf",s);
	  val_label->setLabel(buffer);
	  break;
	case MODE_SCALE_ATOMS:
	  puts("Set Mode: Scale Atoms");
	  mode_label->setLabel("Mode: Scale Atoms");
	  sprintf(buffer,"scale=%5.2lf",sa);
	  val_label->setLabel(buffer);
	  break;
	case MODE_CHANGE_TIME:
	  puts("Set Mode: Change Time");
	  if(maxtime<0.0){
	    timeAtomList.getRange(mintime,maxtime);
	    t=0.0;
	    timeAtomList.getAtoms(t,atomlist);
	  }
	  mode_label->setLabel("Mode: Change Time");
	  sprintf(buffer,"time=%5.2f",t);
	  val_label->setLabel(buffer);
	  break;
	case MODE_NAVIGATE:
	  puts("Set Mode: Navigate");
	  mode_label->setLabel("Mode: Navigate");
	  //mode_label->optimize();
	  navigation->purgeTime(); // purge due to delay in time
	  val_label->setLabel("<>");
	  break;
	}
      }
    }
    else
      button_was_down=0; // button was released after having been pressed

    //------Now manage the current mode
    switch(mode){
      case MODE_SCALE_MODEL:{
      if(THRESH(jy,0.1)){
	if(jy>0)
	  s *= (1.0L+jy*SPEED);
	else 
	  s = s/(1.0L-jy*SPEED);
	// clamp at 100 and 0.1
	if(s>100) s=100;
	if(s<0.1) s=0.1;
	//printf("Model Scale = %lf\n",s);
	modelDCS->setScale((float)s);
	sprintf(buffer,"scale=%5.2lf",s);
	val_label->setLabel(buffer);
      }
    }
      break;
    case MODE_SCALE_ATOMS:{
      if(THRESH(jy,0.1)){
	if(jy>0)
	  sa*= (1.0L+jy*SPEED);
	else 
	  sa= sa/(1.0L-jy*SPEED);
	// clamp at 100 and 0.1
	if(sa>100) sa=100;
	if(sa<0.1) sa=0.1;
	for(int i=0;i<atomGeoms.getSize();i++)
	  (atomGeoms[i])->setScale((float)sa);
	sprintf(buffer,"scale=%5.2lf",sa);
	val_label->setLabel(buffer);
      }
      break;
    }
    case MODE_CHANGE_TIME: {
      if(THRESH(jy,0.1)){
	t+=jy;
	if(t>maxtime) t=maxtime;
	if(t<mintime) t=mintime;
	timeAtomList.getAtoms(t,atomlist);
	for(int i=0;i<atomlist.getSize();i++){
	  (dcslist[i])->setTrans(atomlist[i].p.cart.x,atomlist[i].p.cart.y,atomlist[i].p.cart.z);
	}
	sprintf(buffer,"time=%5.2f",t);
	val_label->setLabel(buffer);
      }
    }
    break;
    case MODE_NAVIGATE:
      navigation->update(); // update the navigation
      navigation->updateDCS();
      break;
    default:
      break;
    }
  } // end of main event loop (terminates when you press ESC key)
  
  StopCave(); // Kill all of the cave subprocesses
  return 0;
}

void StartCave(int *argc,char *argv[],pfDCSPtr &movingdcs,pfDCSPtr &stationarydcs){
  //***************Cave Startup
  pfInit();
  pfCAVEConfig(argc,argv,NULL);
  CAVEFar = 1000;
  pfMultiprocess(PFMP_FORK_CULL);
  pfConfig();
  pfCAVEInitChannels(); 
  
  //Local vars (can only allocate after performer/Cave startup
  pfScene *scene = new pfScene;
  pfGeoState *gstate = new pfGeoState;
  pfLightSource *ls;
  pfSCS *lightSCS;
  pfMatrix mat;
  //**************Create the scene with a DCS at the top
  movingdcs = new pfDCS;
  stationarydcs = new pfDCS;

  gstate->setMode(PFSTATE_ENLIGHTING, PF_ON);
  gstate->setMode(PFSTATE_CULLFACE, PFCF_OFF);
  
  scene->setGState(gstate);
  mat.makeRot(45,1,0,0);
  scene->addChild(lightSCS=new pfSCS(mat));
  lightSCS->addChild(ls=new pfLightSource);
  ls->on();
  scene->addChild(movingdcs); 
  scene->addChild(stationarydcs);
  
  pfChannel *chan=pfCAVEMasterChan();
  chan->setScene(scene);
  pfEarthSky *esky = new pfEarthSky();
  chan->setESky(esky);
  // GLOBAL!!!!
  NoLightGState = new pfGeoState;
  NoLightGState->setMode(PFSTATE_ENLIGHTING,PF_OFF);
}

inline void UpdateCave(){
  pfSync();
  pfCAVEPreFrame();
  pfFrame();
  pfCAVEPostFrame();
}

void StopCave(){
  CAVEHalt();
  pfExit();
}

// Draws a cone at location 0,0,0 in the cave (used for debugging)
pfNode *NewScene(){
  pfDCS *nav_dcs = new pfDCS;
  pfMaterial *material = new pfMaterial;
  pfGeoState *gstate = new pfGeoState;
  pfGeode *geom = new pfGeode;
  pfGeoSet *gset = pfdNewCone(64,pfGetSharedArena());
  pfVec4 *color = new pfVec4;

  material->setShininess(100.0);
  material->setColorMode(PFMTL_BOTH,PFMTL_CMODE_AMBIENT_AND_DIFFUSE);
  gstate->setAttr(PFSTATE_FRONTMTL,material);
  gstate->setAttr(PFSTATE_BACKMTL,material);
  color->set(0.5,0.5,0.2,1.0);
  gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,color,0);
  
  geom->addGSet(gset);
  gset->setGState(gstate);

  return geom;
}

// create new atom geometry with specified color, transparency and refinement
// Higher numbers for refinement makes objects smoother, but also lowers 
// drawing performance since the objects are more complex.
pfDCS *NewAtom(float r,float g,float b,float t,int refinement){ // build the scene graph
  pfDCS *nav_dcs = new pfDCS;
  pfMaterial *material = new pfMaterial;
  pfGeoState *gstate = new pfGeoState;
  pfGeode *geom = new pfGeode;
  pfGeoSet *gset = pfdNewSphere(refinement,pfGetSharedArena());
  pfVec4 *color = new pfVec4;

  material->setShininess(100.0);
  material->setColorMode(PFMTL_BOTH,PFMTL_CMODE_AMBIENT_AND_DIFFUSE);
  gstate->setAttr(PFSTATE_FRONTMTL,material);
  gstate->setAttr(PFSTATE_BACKMTL,material);
  color->set(r,g,b,t);
  gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,color,0);
  
  nav_dcs->addChild(geom);
  geom->addGSet(gset);
  gset->setGState(gstate);
  return nav_dcs;
}
