// pfFlexArrayTmpl.hh
#ifndef __PFFLEXARRAY_TMPL_HH_
#define __PFFLEXARRAY_TMPL_HH_

#include <new.h>
#define PFDOUBLEWORD 8 // sgi doubleword == -align64
#include <Performer/pr.h>
#include <Performer/pr/pfMemory.h>
/*-------------
  Template: pfFlexArray
  Purpose: 
  Same as FlexArray, except it properly manages allocation using
  performer memory allocation routines.  This allows you to use
  performer-allocated arrays as automatic variables (since it is
  performer-allocated internally, the refcount will be handled
  properly when the structure is deleted.

  Needs a copy constructor to fill it out....

  And from the "Purpose" of the FlexArray template:
  Simplifies dynamically allocated array management since it contains
  array size information and manages array resizing.  
  However, since speed-critical
  operations like indexing (the [] operator) are inlined, 
  it adds no speed
  penalty for the most common speed-critical operations.
  Method Summary:
  constructor: The initial "size" is 0 (the size of the array 
  that is accessible
  using the indexing operator [].  However, the internally 
  allocated size is
  a minimum of 8 elements.
  [] : Indexing operator.  Acts just like [] would for a 
  generic array.  Since
  it is inlined, there is actually no penalty in performance 
  compared to
  using a generic array.
  append : Add an element to the end of the array, resizing the 
  array if necessary.
  The arrays are allocated in blocks so it doesn't need to 
  realloc on every append.
  For arrays < 1024 elements long, it doubles the array size 
  on each realloc.  When
  larger than 1024 elements, it adds 1024 elements to the array 
  on every realloc.
  setSize():   Sets the internal and externally availible 
  array sizes to the
  specified size; shrinking a larger internal array if necessary.
  getSize():   Return the current externally availible array size.  
  The internal
  size may be larger.
	
---------------*/

/* This version of the class is designed specifically for performer support
   operator "delete" is not allowable with performer objects */
template<class T>
class pfFlexArray {
  int size,maxsize;
  T *data;
  void *mem;
  // Internal reallocation procedure.
  void grow(int targetSize=1){
    //cerr << "grow: targetSize=" << targetSize << endl;
    do{
      if(!maxsize) maxsize=1;
      if(maxsize<1024)
	maxsize<<=1; // double in size
      else
	maxsize+=1024; // increase by 1k element block
    } while(maxsize<targetSize);
    //cerr << "grow: maxsize=" << maxsize << endl;
    T *newdata;
    void *newmem = pfMalloc(sizeof(T)*maxsize,pfGetSharedArena());
    // newmem = pfMemory::getMemory(newmem);
    newdata = (T*)pfMemory::getData(newmem);
    //T *newdata = new T[sizeof(T)*maxsize];
    // has to do its own copy because realloc is not always compatible with
    // new/delete on all machines
    // I'd like to use pfRealloc() but it isn't clear how to spec arena
    for(int i=0;i<size;i++)
      newdata[i]=data[i];
    pfDelete(mem);
    //cerr << "delete MemAddr    " << (unsigned)(mem) << endl;
    //cerr << "  memdelete returns " << pfDelete(mem) << endl;
    data=newdata;
    mem=pfMemory::getMemory(newmem);
    //cerr << "\tgrow::MemAddr=" << (unsigned)mem << "size=" << maxsize << endl;
  }
public:
	// constructor
  pfFlexArray(int sz=0):size(sz){
    if(sz>PFDOUBLEWORD) 
      maxsize=sz;
    else 
      maxsize=PFDOUBLEWORD;
    mem = pfMalloc(sizeof(T)*maxsize,pfGetSharedArena());
    mem = pfMemory::getMemory(mem);
    data = (T*)pfMemory::getData(mem);
    //cerr << "MemAddr=" << (unsigned)mem << " DataAddr=" << (unsigned)data << 
    //  " RefCount=" << pfGetRef(mem) << endl;
    //data = (T*)pfMalloc(sizeof(T)*maxsize,pfGetSharedArena());
    //data = new T[sizeof(T)*maxsize];
  }
  // destructor
  ~pfFlexArray() //{ if(data) pfDelete(data); }
  {
    pfDelete(mem);
    /* 
    if(pfGetRef(mem)==0) 
      cerr << "pfFlexArray deleting" << pfDelete(mem) << endl;
    else
      cerr << "pfRef for memdelete=" << pfGetRef(mem) << endl;
    */
  }
  inline int getSize(){ return size; }
  int setSize(int s){
    if(s<=maxsize){
      size=s; //don't attempt to shrink array (could have unintended consequences)
      return size; // can't allow shrinkage below current size
    }
    // otherwise if maxsize>s>size>
    maxsize=s; 
    T *newdata;
    void *newmem = pfMalloc(sizeof(T)*maxsize,pfGetSharedArena());
    newdata = (T*)pfMemory::getData(newmem);
    // copy into the new memory
    for(int i=0;i<size && i<maxsize;i++)
      newdata[i]=data[i];
    pfDelete(mem);
    data=newdata;
    mem=pfMemory::getMemory(newmem);
    size=s;
    return size;
  }
  inline int append(T value){
    if(size>=maxsize) // failsafe check
      grow(size+1); // this array only increases in size
    data[size]=value;
    size++;
    return size;
  }
  inline int append(T *values,int nvals){
    if(size+nvals >= maxsize)
    grow(size+nvals);//this array only increases in size
    for(int i=0;i<nvals;i++)
      data[size++]=values[i];
    return size;
  }
  inline T &operator[](int index){ return (data[index]); }
  inline operator T*() { return data; }  // typecasting operator
  inline T *getData() { return data; }   // sometimes you need direct access to the array.
  inline void *getMemory() { return mem; } // get the pfMemory header
  // generally this is only needed when the compiler gets confused by the
  // template and produces bad code (happens occasionally).
};


#endif // __PFFLEXARRAY_TMPL_HH_
